/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is bookmarktools.
 *
 * The Initial Developer of the Original Code is
 * Maik Gollnick <maik.gollnick@gmx.de>.
 * Portions created by the Initial Developer are Copyright (C) 2008
 * the Initial Developer. All Rights Reserved.
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
 
const AUTOMATIC = 0;
const MANUAL = 1;

const DELAY_ONE_WEEK = 604800000;
const DELAY_TWO_WEEK = 1209600000;
const DELAY_THREE_WEEK = 1814400000;
const DELAY_FOUR_WEEK = 2419200000;

const STATUS_SUCCESS = 0;
const STATUS_CORRUPTED = 1;
const STATUS_UNTRACEABLE = 2;

const TYPE_URL = 0;
const TYPE_CONTENT = 1;

/*
	Basic indexer that wraps the clucene component into the js world.	
*/
var Index = {

	path : "",
	delay : DELAY_ONE_WEEK, //default delay one week (in millis)
	mode : MANUAL, //default mode
	filter : null,
	crawler : null,
	
	indexWriter : null,
	indexReader : null,

	//TODO: implement _init()
	init : function(){
		var instWriter = Cc["@bookmarktools.mozdev.org/search/indexWriter;1"].createInstance();
		this.indexWriter = instWriter.QueryInterface(Ci.ICLWriter);
			
		var instReader = Cc["@bookmarktools.mozdev.org/search/indexReader;1"].createInstance();
		this.indexReader = instReader.QueryInterface(Ci.ICLReader);
	
		this.indexReader.setPath(BookmarkTools.utils.fileio.getSearchDir().path); 
		this.setMode(Prefs.getInt("buildType")); 
		
		switch(Prefs.getInt("indexDelay")){
			case 1: this.delay = DELAY_ONE_WEEK; break;
			case 2: this.delay = DELAY_TWO_WEEK; break;
			case 3: this.delay = DELAY_THREE_WEEK; break;
			case 4: this.delay = DELAY_FOUR_WEEK; break;
			default: throw 'Error: Unknown delay!';
		}
	},
	
	setMode : function(mode){
		this.mode = mode;
		switch(this.mode){
			case AUTOMATIC: 
				this.path = BookmarkTools.utils.fileio.getBuildDir().path;
				break;
			case MANUAL:
				this.path = BookmarkTools.utils.fileio.getSearchDir().path;
				break;
			default: throw 'Error: Unknown build type!';
		}
	
		this.indexWriter.setPath(this.path);
	},
	
	setFilter : function(filter){
		this.filter = filter;
		this.filter.onComplete = this.onComplete;
		this.filter.onDone = this.onDone;
		this.filter.onAbort = this.onAbort;
	},
	
	setCrawler : function(crawler){
		this.crawler = crawler;
		this.crawler.onComplete = this.onComplete;
		this.crawler.onDone = this.onDone;
		this.crawler.onAbort = this.onAbort;
	},
	
	onComplete : function(combinedContent, bookmark){
		Index.indexWriter.appendDocument(
			Book.getURI(bookmark),
			Book.getName(bookmark),
			combinedContent
		);
		if(Index.mode == AUTOMATIC){
			var file = Prefs.getFile('alreadyAddedBookmarks');
			BookmarkTools.utils.fileio.writeLineToFile(file, Book.getURI(bookmark) + StringUtils.LF);
		}
	},
	
	onDone : function(){
		if(Index.filter){
			Index.filter.flushDuplicates();
			if(Index.filter.view){
				Index.filter.view._init();
				if(Index.mode == MANUAL) BookmarkTools.panel.index.startLookup();
			}
			Index.filter._init();
		}
		Index.indexWriter.optimize();
		
		if(Index.mode == AUTOMATIC){
			Prefs.setBool('wasUpdated', true);
			Prefs.setBool('firstRun', true);
			Prefs.remove("alreadyAddedBookmarks");
		}
			
		var currentTime = new Date().getTime();
		Prefs.setChar('intIndexDate', currentTime);
	},
	
	onAbort : function(){
		if(Index.filter){
			Index.filter.flushDuplicates();
			if(Index.filter.view){
				Index.filter.view._init();
				if(Index.mode == MANUAL) BookmarkTools.panel.index.startLookup();
			}
			Index.filter._init();
		}
		Index.indexWriter.optimize();
	},
	
	isValidBuildTime : function(){
		if(this.mode == AUTOMATIC){
			var lastBuildTime = 0;
			try {
				lastBuildTime = parseInt(Prefs.getChar('intIndexDate'));
			} catch(e){}
			
			var currentTime = new Date().getTime();
			
			if(isNaN(lastBuildTime)){
				lastBuildTime = 0
			}
	
			if((lastBuildTime + this.delay) <= currentTime){
				return true;
			}
		}
		return false;
	},
	
	// starts building the index
	_build : function(){
		if(this.filter){
			this.filter.go();
		}else {
			if(this.crawler){ //TODO: check -> useful case?
				this.crawler._build();
			}else {
				throw 'Error: No filter available!';
			}
		}
	},

	// finds the pattern and update the given container instance
	_search : function(pattern, container, chunk, chunkSize){
		var max = 0;
		if(this.indexReader.exists()){ //index exists ?
			// validate pattern
			setAlertUser(true);
		
			var query = !doCheckLuceneQueryValue(pattern) ? StringUtils.EMPTY : pattern;
			query = StringUtils.trim(query);

			if(query.length > 2){

				var array = NativeComponents.nsIMutableArray();
				max = this.indexReader.search(query, array, chunk, chunkSize);

				//clear result tree
				container.clear();

				//draw results
				var e = array.enumerate();
				while (e.hasMoreElements()) {
				
					var resultNode = e.getNext().QueryInterface(Ci.IResultNode);
					var nodes = Book.getNode(resultNode.primaryKey);
					
					/** if nodes is empty the url isn't bookmarked */
					if(ArrayUtils.isNotEmpty(nodes)){
					
						var node = nodes[0];
						container.add(
							Book.getName(node), 
							Book.getURI(node),
							Book.getIcon(node), 
							resultNode.score == 0 ? StringUtils.EMPTY : resultNode.score
						);
						
					}else {
						//TODO: string bundle -> '<Untraceable Bookmark>'
						container.add(
							'<Untraceable Bookmark>', 
							resultNode.primaryKey,
							"chrome://bookmarktools/skin/exclamation.png", 
							resultNode.score == 0 ? StringUtils.EMPTY : resultNode.score
						);
					}
				}
				
			}else {
				container.clear();
			}
		}else {
			throw 'Index is empty!';
		}
		return max;
	},
	
	// returns the current index size, that means the number of indexed documents
	_size : function(){
		return this.indexReader.size();
	},
	
	// deletes all or a subset of documents
	//
	//	@param	nodes instance of ExtendedHash
	//
	_delete : function(nodes){

		if(nodes instanceof ExtendedHash){

			if(nodes.length() == this._size() || nodes.isEmpty()){
				
				//it is possible to delete every document but
				//faster to clear the whole directory
				BookmarkTools.utils.fileio.clear(BookmarkTools.utils.fileio.getSearchDir());
				
			}else { //subset

				if(!Prefs.getBool('isRunning')){
					//cleanup
					nodes.each(function(key, value){
						//delete index document
						Index.indexWriter.deleteDocument(value); 
					});
					this.indexWriter.optimize();
				}

			}
			
		}else {
			throw 'Error: Invalid object type (Index.delete).';
		}
		
	},

	// return the index as mutable array
	_lookup : function(){
		var array = NativeComponents.nsIMutableArray();
		this.indexReader.lookup(array);
		return array;
	},
	
	// if build is running this function will abort the crawler
	_abort : function(){
		if(!this.filter) throw 'Error: No filter available!';
		if(!this.filter.crawler) throw 'Error: No crawler available!';
		
		this.filter.crawler._abort();
		// if indexer was running reset this flag
		if(Prefs.getBool('isRunning'))
			Prefs.setBool('isRunning', false);
	}

};

Index.init(); 

var Filter = {

	uniqueContent : null,
	duplicateSelector : null,
	
	countDuplicate : 0,
	countError : 0,
	count : 0,
	max : 0,
	
	crawler : null,
	view : null,
	
	check : false,
	
	logger : null,
	
	init : function(){
		Parser.init();
		this.uniqueContent = [];
		this.duplicateSelector = new ExtendedHash();
		this.logger = new Logger(this, arguments);
		this.logger.init("filter.log");
	},
	
	_init : function(){
		Parser._init();
		this.uniqueContent = null;
		this.duplicateSelector = null;
		this.countDuplicate = 0;
		this.countError = 0;
		this.count = 0;
		this.max = 0;
		this.check = false;
	},
	
	setCrawler : function(crawler, count, max){
		this.crawler = crawler;
		this.crawler.onComplete = this.filter;
		this.crawler.onError = this.updateError;
		this.crawler.onDone = this.onDone;
		this.crawler.onAbort = this.onAbort;
		
		if(arguments.length == 3){
			this.count = count;
			this.max = max;
		}else {
			this.max = this.crawler.bookmarks.length();
		}
	},
	
	getBookmarksFromFile : function(){
		var file = Prefs.getFile('alreadyAddedBookmarks');
		var alreadyAddedBookmarks = BookmarkTools.utils.fileio.readFileLineByLine(file);
	
		var bookmarks = Book.getResourceHash();
		var notAddedBookmarks = new ExtendedHash();
		
		if(alreadyAddedBookmarks.isEmpty()){
			return {
				currentIndexed : 0,
				allBookmarks : bookmarks.length(),
				remainingBookmarks : bookmarks
			};
		}

		bookmarks.each(function(key, value){
			if(!alreadyAddedBookmarks.contains(key)){
				notAddedBookmarks.put(key, value);
			}
		});
		
		return {
			currentIndexed : alreadyAddedBookmarks.length(),
			allBookmarks : bookmarks.length(),
			remainingBookmarks : notAddedBookmarks
		};
	},
	
	flushDuplicates : function(){
		Filter.duplicateSelector.each(function(key, value){
			Filter.updateDuplicate(key, value);
		});
	},
	
	onComplete : function(combinedContent, bookmark){},
	
	onDone : function(){
		Filter.flushDuplicates();
		if(Filter.view)
			Filter.view._init();
		Filter._init();
	},
	
	onAbort : function(){
		Filter.flushDuplicates();
		if(Filter.view)
			Filter.view._init();
		Filter._init();
	},
	
	setView : function(view){
		this.view = view;
	},
	
	setCheck : function(check){
		this.check = check;
	},
	
	go : function(){
		if(!this.crawler) throw 'Error: No crawler available!';
		
		//url duplicates loaded one time at startup in automatic build mode
		if((Index.mode == AUTOMATIC && Prefs.getBool('firstRun')) ||
			Index.mode == MANUAL){
			var duplicateURLs = this.crawler.bookmarks.trash();
			for(var i = 0, l = duplicateURLs.length ;i < l; i++){
				
				var node = duplicateURLs[i];
				var uri = Book.getURI(node);
				
				if(Filter.duplicateSelector.contains(uri)){
					Filter.duplicateSelector.get(uri).push({bookmark: node, type: TYPE_URL});
				}else {
					Filter.duplicateSelector.put(uri, [{bookmark: node, type: TYPE_URL}]);
				}
			}
		}
		
		//update counter because automatic indexing
		//is done in 0-n steps and every step produces duplicates
		if(Index.mode == AUTOMATIC){
			this.countDuplicate = Prefs.getBranch('dupBm').length;
			this.countError = Prefs.getBranch('invBm').length;
			
			// BrowserView does not need any uri, so setProgress(..) has only 2 arguments
			if(Filter.view && Filter.view == BrowserView) 
				Filter.view.setProgress(Filter.count, Filter.max);
			else throw 'Error: Invalid view type for automatic mode!'
		}
	
		this.crawler._build();
	},
	
	//overwrites crawlers onNext and onUpdate
	filter : function(responseText, mimeType, charSet, bookmark){
		var content = StringUtils.EMPTY;
		try{
			content = Parser.parse(mimeType, charSet, responseText);
		}catch(e){
			alert(e);
		}

		var uri = Book.getURI(bookmark);
		Filter.logger.info(uri, this, arguments.callee);
		
		if(!Filter.isDuplicate(content, bookmark)){
			if(!Filter.check){
				var combinedContent = Filter.combineContent(content, bookmark);
				Filter.onComplete(combinedContent, bookmark);
			}	
		}
	
		Filter.count++;
		if(Filter.view)
			Filter.view.setProgress(Filter.count, Filter.max, uri);
	},
	
	combineContent : function(content, bookmark){
		content += StringUtils.SPACE;
		content += Book.getName(bookmark);
		content += StringUtils.SPACE;
		content += Book.getURI(bookmark);
		content += StringUtils.SPACE;
		content += Book.getDesc(bookmark);
		content += StringUtils.SPACE;
		content += Book.getKeys(bookmark);
		return StringUtils.trim(content);
	},
	
	combineWithoutContent : function(bookmark){
		var combined = Book.getName(bookmark);
		combined += StringUtils.SPACE;
		combined += Book.getURI(bookmark);
		combined += StringUtils.SPACE;
		combined += Book.getDesc(bookmark);
		combined += StringUtils.SPACE;
		combined += Book.getKeys(bookmark);
		return StringUtils.trim(combined);
	},
	
	trimContent : function(content){
		var trimed = StringUtils.stripScripts(content);
		trimed = Parser.removeWhiteSpaces(trimed);
		return Parser.removeNumbers(trimed);
	},
	
	isDuplicate : function(content, node){
		var trimedContent = Filter.trimContent(content);
	
		//if no content is available continue indexing to
		//appends all bookmark properties in every case
		if(StringUtils.isEmpty(trimedContent)) return false;
		
		if(!Filter.uniqueContent[trimedContent]){
			Filter.uniqueContent[trimedContent] = Book.getURI(node);
			return false;
		}
		
		var original = Filter.uniqueContent[trimedContent];
		if(this.duplicateSelector.contains(original)){
			this.duplicateSelector.get(original).push({bookmark: node, type: TYPE_CONTENT});
		}else {
			this.duplicateSelector.put(original, [{bookmark: node, type: TYPE_CONTENT}]);
		}
	
		return true;
	},

	updateDuplicate : function(uri, equals){
		if(Prefs.getBool('cleanupDuplicates')){
		
			for(var i = 0, l = equals.length ;i < l; i++){
				Book.removeBookmark(equals[i].bookmark);
			}
	
		}else {
			for(var i = 0, l = equals.length ;i < l; i++){
				var json = JSONUtils.toJSON(
					{
						uri: Book.getURI(equals[i].bookmark),
						type: equals[i].type
					}
				);
				Prefs.setChar('dupBm' + Filter.countDuplicate++, json);
				
				var duplicate_tree = $('DUPLICATETREE');
				if(duplicate_tree){
					duplicate_tree.add(
						Book.getName(equals[i].bookmark),
						Book.getURI(equals[i].bookmark),
						Book.getIcon(equals[i].bookmark),
						'dupBm' + (Filter.countDuplicate - 1)
					);
					var itemCount = $('DUPLICATE_COUNT');
					itemCount.value = '(' + duplicate_tree.size + ')';
				}
			}
		}
	},
	
	updateError : function(status, statusText, bookmark){
		Filter.count++;
		var json = JSONUtils.toJSON(
			{
				uri: Book.getURI(bookmark),
				status: status,
				statusText: statusText
			}
		);
		
		Filter.logger.info(Book.getURI(bookmark), this, arguments.callee);
		Prefs.setChar('invBm' + Filter.countError++, json);
	
		var error_tree = $('ERRORTREE');
		var other_tree = $('OTHERTREE');
		
		if((error_tree || other_tree) && typeof BookmarkTools.panel.errors.updateView != 'undefined')
			BookmarkTools.panel.errors.updateView(status, statusText, bookmark, Prefs.getInt('errorView'));
		
		//if error type equals 'OTHER' the bookmark is valid but no content available
		//-> in this case only the bookmark properties combined for index document
		if(status <= -400){
			if(!Filter.check){
				var combined = Filter.combineWithoutContent(bookmark);
				Filter.onComplete(combined, bookmark);
			}	
		}
	}

};

var IndexLock = {

	acquire : function(){
		var key = new Date().getTime();
		var lock = StringUtils.EMPTY;

		var successfullyQuit = true;
		try {
			successfullyQuit = Prefs.getBool('successfullyQuit');
		} catch(e) {}

		if(successfullyQuit){
			lock = Prefs.getChar('indexLock');
		}
		
		if(StringUtils.isEmpty(lock)){
			Prefs.setChar('indexLock', key);
			return parseInt(key);
		}
		
		return -1;
	},
	
	release : function(key){
		var lock = Prefs.getChar('indexLock');
		if(key == lock){
			Prefs.setChar('indexLock', StringUtils.EMPTY);
		}
	}

};

var View = {

	doc : null,
	win : null,
	panel : null,
	
	init : function(){
		var w = NativeComponents.nsIWindowMediator('bookmarkTools-manual-indexing');
		View.win = w[0];
		View.doc = w[0].document;
		View.panel = {
			label: View.doc.getElementById('bookmarkTools.progress.info'),
			progressLabel: View.doc.getElementById('bookmarkTools.progress.label'),
			progressBar: View.doc.getElementById('bookmarkTools.progress.bar')
		};
	},
	
	_init : function(){
		View.win.close();
		View.win = null;
		View.doc = null;
		View.panel = null;
	},

	percent : function(i, j){
		return Math.floor((i / j) * 100);
	},
	
	formatURI : function(i, j, uri, maxLength){
		var range = '[' + i + '/' + j + '] ';
		return range + StringUtils.truncate(uri, maxLength, ' [..]');
	},
	
	formatProgress : function(percent){
		return percent + '%';
	},
	
	setProgress : function(i, j, uri){
		View.panel.label.value = View.formatURI(i, j, uri, 30);
		var val = View.percent(i, j);
		View.panel.progressLabel.value = View.formatProgress(val);
		View.panel.progressBar.value = val;
	}

};

var BrowserView = {
	
	doc : null,
	win : null,
	panel : null,
	
	init : function(){
		BrowserView.win = NativeComponents.nsIBrowserWindow();
		BrowserView.doc = BrowserView.win.document;
		BrowserView.panel = {
			progressLabel: BrowserView.doc.getElementById('bookmarkTools.autoprogress.label'),
			progressBar: BrowserView.doc.getElementById('bookmarkTools.autoprogress.bar')
		};
		$('bookmarkTools.autoProgressOverlay').hidden = false;
	},
	
	_init : function(){
		$('bookmarkTools.autoProgressOverlay').hidden = true;
	},
	
	percent : function(i, j){
		return Math.floor((i / j) * 100);
	},
	
	formatProgress : function(percent){
		return percent + '%';
	},
	
	setProgress : function(i, j){
		var val = BrowserView.percent(i, j);
		BrowserView.panel.progressLabel.value = BrowserView.formatProgress(val);
		BrowserView.panel.progressBar.value = val;
	}
};
